from pyphone import data
import os
import json
import re
import scipy.cluster.hierarchy as hierarchy

phone_memory = {}

bad_phones = {
    "ɝ": "ɜ˞",
}


def _get_vector(phones):
    global phone_memory
    if bad_phones.get(phones):
        phones = bad_phones.get(phones)
    try:
        if not phones:
            raise ValueError("No phone passed in")
        if len(phones) == 1:
            return data.phone_features[phones]
        else:
            if phone_memory.get(phones):
                return phone_memory[phones]
            vec = [None] * data.NUM_FEATURES
            # Combine the feature vectors
            mods = []
            for i in range(data.NUM_FEATURES):
                val = None
                for phone in phones:
                    if phone in data.modifiers:
                        mods.append(data.modifiers[phone])
                    else:
                        p_val = data.phone_features[phone][i]
                        if p_val is True:
                            val = True
                            break
                        elif p_val is False:
                            val = False
                vec[i] = val
            for mod in mods:
                for i, v in mod.items():
                    vec[i] = v
            phone_memory = {phones: vec}
            return vec
    except KeyError as e:
        raise ValueError("Unrecognized phone %s" % str(e))


def is_vowel(phone):
    vec = _get_vector(phone)
    return vec[0]


def distance(phoneA, phoneB):
    weights = data.feature_weights
    # Handle spaces specially
    if phoneA == ' ' or phoneB == ' ':
        if phoneA == phoneB:
            # Certain
            return 0
        else:
            # Impossible to sub for something else
            return 1
    if is_vowel(phoneA) != is_vowel(phoneB):
        # If one is a vowel and the other is a consonant
        return 1

    vowel = is_vowel(phoneA)
    if vowel:
        total_weights = data.total_vowel_weight
        features = data.vowel_features
        factor = 1
    else:
        total_weights = data.total_consonant_weight
        features = data.consonant_features
        factor = 2

    vecA = _get_vector(phoneA)
    vecB = _get_vector(phoneB)

    different = 0
    for i in features:
        if vecA[i] != vecB[i]:
            different += weights[i]
    return different / total_weights * factor


# Change this filter to expand to other languages
alpha_only = re.compile('[^a-zA-Z]')


def tokenize(word):
    word = alpha_only.sub('', word)
    return "".join(["_" + letter for letter in word])


def strip_repeats(word):
    stripped = []
    last_letter = None
    word = filter(None, word.split("$"))
    for letter in word:
        if letter != last_letter and not letter.startswith('_'):
            stripped.append(letter)
        last_letter = letter
    return tuple(stripped)


def phoneticize(word, language="english"):
    tokenized = tokenize(word.lower())
    next_words = set([tokenized])
    results = set()
    mappings = get_mapping(language)
    tokenized_mappings = {
        tokenize(key): list(map(lambda x: "$" + x, mapping))
        for (key, mapping) in mappings.items()
    }
    while len(next_words) > 0:
        next_word = next_words.pop()

        has_replacements = False
        for key, replacements in tokenized_mappings.items():
            if key in next_word:
                has_replacements = True
                for replacement in replacements:
                    next_words.add(next_word.replace(key, replacement, 1))
        if not has_replacements:
            results.add(strip_repeats(next_word))
    return results


cached_json = {}


def get_mapping(language):
    cached = cached_json.get(language)
    if cached:
        return cached
    with open(os.path.join(os.path.dirname(__file__),
                           "data",
                           "mappings",
                           language + ".json")) as f:
        mappings = json.load(f)
        cached_json[language] = mappings
        return mappings


cached_phone_sets = {}


def phone_set(language="english"):
    cached = cached_phone_sets.get(language)
    if cached:
        return cached

    phone_set = set()
    mapping = get_mapping(language)
    for key, phones in mapping.items():
        for phone in phones:
            phone_set.add(phone)
    cached_phone_sets[language] = phone_set
    return phone_set


def phonex(word, language="english"):
    """
    Short for phone index, maps a word onto a sequence of phone clusters
    Strips non supported characters (for english, non alpha characters)
    """
    phone_variants = phoneticize(word)
    mappings = cluster_phones(language)
    results = []
    for phone_variant in phone_variants:
        try:
            phonex_variant = tuple([mappings[phone] for phone in
                                    phone_variant])
            results.append(phonex_variant)
        except:
            print('Error:', word, phone_variant)
            exit(1)
    return results


CLUSTER_SENSITIVITY = 0.5
cluster_cache = {}


def cluster_phones(language):
    if cluster_cache.get(language):
        return cluster_cache[language]
    phones = sorted(list(phone_set(language)))
    phones = sorted(phones)
    triangle_distance = []

    for a in range(len(phones)):
        for b in range(a + 1, len(phones)):
            triangle_distance.append(distance(phones[a], phones[b]))

    linkages = hierarchy.linkage(triangle_distance)
    clusters = hierarchy.fcluster(linkages, CLUSTER_SENSITIVITY)
    cluster_map = {}

    for index in range(len(clusters)):
        cluster_map[phones[index]] = clusters[index]

    cluster_cache[language] = cluster_map
    return cluster_map

# import pprint
# pp = pprint.PrettyPrinter(indent=4)
    # for index in range(len(clusters)):
    #     set_phones = forward_map.get(clusters[index])
    #     if not set_phones:
    #         set_phones = []
    #         forward_map[clusters[index]] = set_phones
    #     set_phones.append(phones[index])



# flake8: noqa

# Indexes
# 0 - syllabic
# 1 - sonorant
# 2 - consonantal
# 3 - continuant
# 4 - delayed release
# 5 - lateral
# 6 - nasal
# 7 - strident
# 8 - voice
# 9 - spread glottis
# 10 - constricted glottis
# 11 - anterior
# 12 - coronal
# 13 - distributed
# 14 - labial
# 15 - high (vowel/consonant, not tone)
# 16 - low (vowel/consonant, not tone)
# 17 - back
# 18 - round
# 19 - tense
# 20 - long

NUM_FEATURES = 21

phone_features = {
    "c": [False, False, True, False, False, False, False, None, False, False, False, False, False, None, False, True, False, False, False, None, False],
    "ɡ": [False, False, True, False, False, False, False, None, True, False, False, False, False, None, False, True, False, True, False, None, False],
    "k": [False, False, True, False, False, False, False, None, False, False, False, False, False, None, False, True, False, True, False, None, False],
    "q": [False, False, True, False, False, False, False, None, False, False, False, False, False, None, False, False, False, True, False, None, False],
    "ɖ": [False, False, True, False, False, False, False, None, True, False, False, False, True, False, False, False, False, False, False, None, False],
    "ɟ": [False, False, True, False, False, False, False, None, True, False, False, False, False, None, False, True, False, False, False, None, False],
    "ɠ": [False, False, True, False, False, False, False, None, True, False, True, False, False, None, False, True, False, True, False, None, False],
    "ɢ": [False, False, True, False, False, False, False, None, True, False, False, False, False, None, False, False, False, True, False, None, False],
    "ʄ": [False, False, True, False, False, False, False, None, True, False, True, False, False, None, False, True, False, False, False, None, False],
    "ʈ": [False, False, True, False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, None, False],
    "ʛ": [False, False, True, False, False, False, False, None, True, False, True, False, False, None, False, False, False, True, False, None, False],
    "b": [False, False, True, False, False, False, False, None, True, False, False, True, False, None, True, False, False, False, False, None, False],
    "d": [False, False, True, False, False, False, False, False, True, False, False, True, True, False, False, False, False, False, False, None, False],
    "p": [False, False, True, False, False, False, False, None, False, False, False, True, False, None, True, False, False, False, False, None, False],
    "t": [False, False, True, False, False, False, False, False, False, False, False, True, True, False, False, False, False, False, False, None, False],
    "ɓ": [False, False, True, False, False, False, False, None, True, False, True, True, False, None, True, False, False, False, False, None, False],
    "ɗ": [False, False, True, False, False, False, False, False, True, False, True, True, True, False, False, False, False, False, False, None, False],
    "x": [False, False, True, True, False, False, False, None, False, False, False, False, False, None, False, True, False, True, False, None, False],
    "ç": [False, False, True, True, False, False, False, None, False, False, False, False, False, None, False, True, False, False, False, None, False],
    "ħ": [False, False, True, True, False, False, False, None, False, False, False, False, False, None, False, False, True, True, False, None, False],
    "ɣ": [False, False, True, True, False, False, False, None, True, False, False, False, False, None, False, True, False, True, False, None, False],
    "ʁ": [False, False, True, True, False, False, False, None, True, False, False, False, False, None, False, False, False, True, False, None, False],
    "ʂ": [False, False, True, True, False, False, False, True, False, False, False, False, True, False, False, False, False, False, False, None, False],
    "ʃ": [False, False, True, True, False, False, False, True, False, False, False, False, True, True, False, False, False, False, False, None, False],
    "ʐ": [False, False, True, True, False, False, False, True, True, False, False, False, True, False, False, False, False, False, False, None, False],
    "ʒ": [False, False, True, True, False, False, False, True, True, False, False, False, True, True, False, False, False, False, False, None, False],
    "ʕ": [False, False, True, True, False, False, False, None, True, False, False, False, False, None, False, False, True, True, False, None, False],
    "ʝ": [False, False, True, True, False, False, False, None, True, False, False, False, False, None, False, True, False, False, False, None, False],
    "χ": [False, False, True, True, False, False, False, None, False, False, False, False, False, None, False, False, False, True, False, None, False],
    "f": [False, False, True, True, False, False, False, None, False, False, False, True, False, None, True, False, False, False, False, None, False],
    "s": [False, False, True, True, False, False, False, True, False, False, False, True, True, False, False, False, False, False, False, None, False],
    "v": [False, False, True, True, False, False, False, None, True, False, False, True, False, None, True, False, False, False, False, None, False],
    "z": [False, False, True, True, False, False, False, True, True, False, False, True, True, False, False, False, False, False, False, None, False],
    "ð": [False, False, True, True, False, False, False, False, True, False, False, True, True, True, False, False, False, False, False, None, False],
    "ɸ": [False, False, True, True, False, False, False, None, False, False, False, True, False, None, True, False, False, False, False, None, False],
    "β": [False, False, True, True, False, False, False, None, True, False, False, True, False, None, True, False, False, False, False, None, False],
    "θ": [False, False, True, True, False, False, False, False, False, False, False, True, True, True, False, False, False, False, False, None, False],
    "ɧ": [False, False, True, True, True, False, False, None, False, False, False, False, True, True, False, True, False, None, False, None, False],
    "ɕ": [False, False, True, True, True, False, False, True, False, False, False, True, True, True, False, True, False, False, False, None, False],
    "ɬ": [False, False, True, True, True, True, False, False, False, False, False, True, True, False, False, None, None, None, False, None, False],
    "ɮ": [False, False, True, True, True, True, False, False, True, False, False, True, True, False, False, None, None, None, False, None, False],
    "ʑ": [False, False, True, True, True, False, False, True, True, False, False, True, True, True, False, True, False, False, False, None, False],
    "ɱ": [False, True, True, False, None, False, True, None, True, False, False, True, False, None, True, None, None, None, False, None, False],
    "ʔ": [False, True, False, False, False, False, False, None, False, False, True, False, False, None, False, False, False, False, False, None, False],
    "ŋ": [False, True, True, False, False, False, True, None, True, False, False, False, False, None, False, True, False, True, False, None, False],
    "ɳ": [False, True, True, False, False, False, True, False, True, False, False, False, True, None, False, False, False, False, False, None, False],
    "ɴ": [False, True, True, False, False, False, True, None, True, False, False, False, False, None, False, False, False, True, False, None, False],
    "m": [False, True, True, False, False, False, True, None, True, False, False, True, False, None, True, False, False, False, False, None, False],
    "n": [False, True, True, False, False, False, True, False, True, False, False, True, True, False, False, False, False, False, False, None, False],
    "ɲ": [False, True, True, False, False, False, True, False, True, False, False, True, False, None, False, True, False, False, False, None, False],
    "ɥ": [False, True, False, True, None, False, False, None, True, False, False, None, False, None, True, True, False, False, True, True, False],
    "ɰ": [False, True, False, True, None, False, False, None, True, False, False, None, False, None, False, True, False, None, False, True, False],
    "ʋ": [False, True, False, True, None, False, False, None, True, False, False, True, False, None, True, None, None, None, False, None, False],
    "ʀ": [False, True, True, True, None, False, False, None, True, False, False, None, False, None, False, False, False, True, False, None, False],
    "ʙ": [False, True, True, True, None, False, False, None, True, False, False, True, False, None, True, None, None, None, False, None, False],
    "ʟ": [False, True, True, True, None, True, False, None, True, False, False, None, False, None, False, True, False, None, False, None, False],
    "ɭ": [False, True, True, True, None, True, False, False, True, False, False, False, True, False, False, None, None, None, False, None, False],
    "ɽ": [False, True, True, True, None, False, False, False, True, False, False, False, True, False, False, None, None, None, False, None, False],
    "ʎ": [False, True, True, True, None, True, False, None, True, False, False, False, True, True, False, True, False, False, False, None, False],
    "r": [False, True, True, True, None, False, False, False, True, False, False, True, True, False, False, None, None, None, False, None, False],
    "ɫ": [False, True, True, True, None, True, False, False, True, False, False, True, True, False, False, False, False, True, False, None, False],
    "ɺ": [False, True, True, True, None, True, False, False, True, False, False, True, True, False, False, None, None, None, False, None, False],
    "ɾ": [False, True, True, True, None, False, False, False, True, False, False, True, True, False, False, None, None, None, False, None, False],
    "ʍ": [False, True, False, True, False, False, False, None, False, False, False, False, False, None, True, True, False, True, True, None, False],
    "h": [False, True, True, True, False, False, False, False, False, False, False, False, False, None, False, False, False, False, False, None, False],
    "j": [False, True, False, True, False, False, False, None, True, False, False, False, False, None, False, True, False, False, False, None, False],
    "w": [False, True, False, True, False, False, False, None, True, False, False, False, False, None, True, True, False, True, True, None, False],
    "ɹ": [False, True, False, True, False, False, False, False, True, False, False, False, True, False, False, True, False, True, True, None, False],
    "ɻ": [False, True, False, True, False, False, False, False, True, False, False, False, True, False, False, False, False, False, False, None, False],
    "l": [False, True, True, True, False, True, False, False, True, False, False, True, True, False, False, False, False, False, False, None, False],
    "ɦ": [False, True, True, True, False, False, False, None, False, False, False, False, False, None, False, False, False, False, False, None, False],
    "ɑ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, False, False, True, True, False, True, False],
    "ɘ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, False, False, False, False, False, True, False],
    "ɞ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, True, False, False, False, True, False, False],
    "ɤ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, False, False, False, True, False, True, False],
    "ɵ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, True, False, False, False, True, True, False],
    "ʉ": [True, True, False, True, None, False, False, None, True, False, False, False, False, False, True, True, False, False, True, True, False],
    "a": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, True, True, False, True, False],
    "e": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, False, False, True, False],
    "i": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, False, False, True, False],
    "o": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, True, True, False],
    "u": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, True, True, False, True, True, True, False],
    "y": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, True, True, False, False, True, True, False],
    "æ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, True, False, False, True, False],
    "ø": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, False, True, True, False],
    "œ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, False, True, False, False],
    "ɒ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, True, True, True, True, False],
    "ɔ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, True, False, False],
    "ə": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, False, False, False],
    "ɘ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, False, False, False],
    "ɵ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, True, False, False],
    "ɞ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, True, True, False],
    "ɜ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, False, True, False],
    "ɛ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, False, False, False, False],
    "ɨ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, True, False, True, False],
    "ɪ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, False, False, False, False],
    "ɯ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, True, False, False, False],
    "ɶ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, True, False, True, True, False],
    "ʊ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, True, True, False, False],
    "ɐ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, False, True, False],
    "ʌ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, False, False, True, False, True, False],
    "ʏ": [True, True, False, True, False, False, False, None, True, False, False, False, False, False, False, True, False, False, True, False, False],
}

vowel_features = set([0, 1, 18, 3, 19, 17, 8, 15, 14, 16, 18, 20])
consonant_features = set([1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19])


modifiers = {
    "ː": {
        20: True,
    },
    " ̻": {
        13: True,
    },
    "ˤ": {
        16: True,
        17: True
    },
    "˞": {
        11: False,
        15: True,
        18: True
    }
}

feature_weights = [0.14285714285714285, 0.14285714285714285, 0.14285714285714285, 0.07142857142857142, 0.03571428571428571, 0.03571428571428571, 0.03571428571428571, 0.017857142857142856, 0.017857142857142856, 0.017857142857142856, 0.017857142857142856, 0.03571428571428571, 0.03571428571428571, 0.017857142857142856, 0.03571428571428571, 0.03571428571428571, 0.03571428571428571, 0.03571428571428571, 0.03571428571428571, 0.03571428571428571, 0.017857142857142856]
total_vowel_weight = sum([feature_weights[vowel_feature] for vowel_feature in vowel_features])
total_consonant_weight = sum([feature_weights[consonant_feature] for consonant_feature in consonant_features])

# vowel_weights = [1 / (len(vowel_features) - 1) if i in vowel_features[1:] else 0 for i in range(21)]
# consonant_weights = [1 / len(consonant_features) if i in consonant_features else 0 for i in range(21)]
standard_weights = [1 / 21 for i in range(21)]
zero_weights = [0 for i in range(21)]



#!/usr/bin/env python3

from distutils.core import setup
setup(
  name = 'pyphone',
  packages = ['pyphone'], # this must be the same as the name above
  version = '0.1.5',
  description = 'Phonetic manipulation library',
  author = 'Ling Zhang',
  author_email = 'lz@ling.nz',
  url = 'https://github.com/lingz/pyphone', # use the URL to the github repo
  install_requires = [
    'scipy'
  ],
  include_package_data = True,
  keywords = ['nlp', 'phonetic', 'ipa'],
    package_data = {
        'pyphone': ['data/mappings/*.json']
    }
)

